@@ -28,6 +28,9 @@ from lxml import etree
 #
 # Globals and constants
 
+Namespaces = {'xs': 'http://www.w3.org/2001/XMLSchema'}
+Xsd_namespace_uri = 'http://www.w3.org/2001/XMLSchema'
+
 
 #
 # Functions for external use
@@ -58,6 +61,8 @@ class Params(object):
 class SchemaIOError(IOError):
     pass
 
+class RaiseComplexTypesError(Exception):
+    pass
 
 #
 # Functions for internal use and testing
@@ -69,6 +74,7 @@ def clear_includes_and_imports(node):
     child_iter2 = node.iterfind('{%s}import' % (namespace, ))
     for child in itertools.chain(child_iter1, child_iter2):
         repl = etree.Comment(etree.tostring(child))
+        repl.tail = '\n'
         node.replace(child, repl)
 
 
@@ -156,6 +162,7 @@ def collect_inserts_aux(child, params, inserts, options):
                 if (child1.tag != '{%s}include' % (namespace, ) and
                     child1.tag != '{%s' % (namespace, )):
                     comment = etree.Comment(etree.tostring(child))
+                    comment.tail = '\n'
                     inserts.append(comment)
                     inserts.append(child1)
         collect_inserts(root, params, inserts, options)
@@ -186,6 +193,7 @@ def prep_schema_doc(infile, outfile, inpath, options):
     for insert_node in inserts:
         root2.append(insert_node)
     process_groups(root2)
+    raise_anon_complextypes(root2)
     doc2 = etree.ElementTree(root2)
     doc2.write(outfile)
     return doc2
@@ -208,12 +216,13 @@ def prep_schema(inpath, outpath, options):
     if outpath:
         outfile.close()
 
-Namespaces = {'xs': 'http://www.w3.org/2001/XMLSchema'}
 
 def process_groups(root):
+    # Get all the xs:group definitions at top level.
     defs = root.xpath('./xs:group', namespaces=Namespaces)
     defs = [node for node in defs if node.get('name') is not None]
-    refs = root.xpath('.//xs:group', namespaces=Namespaces)
+    # Get all the xs:group references (below top level).
+    refs = root.xpath('./*//xs:group', namespaces=Namespaces)
     refs = [node for node in refs if node.get('ref') is not None]
     # Create a dictionary of the named model groups (definitions).
     def_dict = {}
@@ -225,6 +234,8 @@ def process_groups(root):
 def replace_group_defs(def_dict, refs):
     for ref_node in refs:
         name = trim_prefix(ref_node.get('ref'))
+        if name is None:
+            continue
         def_node = def_dict.get(name)
         if def_node is not None:
             content = def_node.xpath('./xs:sequence|./xs:choice|./xs:all',
@@ -233,10 +244,71 @@ def replace_group_defs(def_dict, refs):
                 content = content[0]
                 parent = ref_node.getparent()
                 for node in content:
-                    ref_node.addprevious(deepcopy(node))
+                    new_node = deepcopy(node)
+                    # Copy minOccurs and maxOccurs attributes to new node.
+                    value = ref_node.get('minOccurs')
+                    if value is not None:
+                        new_node.set('minOccurs', value)
+                    value = ref_node.get('maxOccurs')
+                    if value is not None:
+                        new_node.set('maxOccurs', value)
+                    ref_node.addprevious(new_node)
                 parent.remove(ref_node)
 
 
+def raise_anon_complextypes(root):
+    """ Raise each anonymous complexType to top level and give it a name.
+    Rename if necessary to prevent duplicates.
+    """
+    element_tag = '{%s}element' % (Xsd_namespace_uri, )
+    def_names = {}
+    # Collect top level complexTypes.
+    defs = root.xpath('./xs:complexType', namespaces=Namespaces)
+    for node in defs:
+        type_name = node.get('name')
+        def_names[type_name] = node
+    # Collect top level simpleTypes.
+    defs = root.xpath('./xs:simpleType', namespaces=Namespaces)
+    for node in defs:
+        type_name = node.get('name')
+        def_names[type_name] = node
+    # Find all complexTypes below top level.
+    #   Raise them to top level and name them.
+    #   Re-name if there is a duplicate (simpleType, complexType, or
+    #   previous renamed type).
+    #   Change the parent (xs:element) so the "type" attribute refers to 
+    #   the raised and renamed type.
+    #   Collect the new types.
+    el = etree.Comment(text="Raised anonymous complexType definitions")
+    el.tail = "\n\n"
+    root.append(el)
+    defs = root.xpath('./*/*//xs:complexType', namespaces=Namespaces)
+    for node in defs:
+        parent = node.getparent()
+        if parent.tag != element_tag:
+            continue
+        name = parent.get('name')
+        if not name:
+            continue
+        type_name = '%sType' % (name, )
+        type_name = unique_name(type_name, def_names)
+        def_names[type_name] = node
+        parent.set('type', type_name)
+        node.set('name', type_name)
+        # Move the complexType node to top level.
+        root.append(node)
+
+def unique_name(type_name, def_names):
+    orig_type_name = type_name
+    count = 0
+    while count < 100:
+        if type_name not in def_names:
+            return type_name
+        count += 1
+        type_name = '%s%d' % (orig_type_name, count, )
+    raise RaiseComplexTypesError('duplicate name count max (100) exceeded')
+
+
 def trim_prefix(name):
     names = name.split(':')
     if len(names) == 1:
